/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <QtGui/QApplication>
#include <QtGui/QMenuBar>
#include <QtGui/QFileDialog>
#include <QtGui/QMessageBox>
#include <QtGui/QToolBar>
#include <QtGui/QStatusBar>

#include "boost/foreach.hpp"

#include "wscout_gui_main_window.hh"
#include "wscout_gui_filter_dialog.hh"
#include "wscout_gui_pixmap.hh"
#include "wscout_pcap_trace.hh"
#include "wscout_config.hh"

/// WScout constructs.
namespace wscout
{

  /// WScout's GUI elements.
  namespace gui
  {

    MainWindow::MainWindow(): focus_finder_ (QKeySequence::Find, this)
    {
      initWidgets();
      updateTrace();
    }

    MainWindow::MainWindow(const MainWindow& w):
      focus_finder_ (QKeySequence::Find, this)
    {
      initWidgets();

      if (const int c = w.tabs_.count())
	for (int i = 0; i < c; ++i)
	  {
	    const Trace* const t =
	      dynamic_cast<const Trace*> (w.tabs_.widget(i));

	    assert(t and t->trace());
	    openTrace(t->trace()->file_name().c_str());
	  }
      else
	updateTrace();
    }

    bool
    MainWindow::open()
    {
      Trace* const t = trace();

      QString dir = t ?
	QFileInfo (t->trace()->file_name().c_str()).absolutePath():
	QDir::currentPath();

      QString	traceName =
	QFileDialog::getOpenFileName(this, tr("Open trace"), dir);

      if (traceName.isNull())
	return false;
      return openTrace(traceName);
    };

    bool
    MainWindow::openTrace(const QString& traceName)
    {
      pcap::Trace*		pcapTrace = 0;
      QProgressDialog*		dialog = 0;
      NewProgressDialog		setup (&dialog, this);
      UpdateProgressDialog	update (&dialog);

      try
        {
          pcapTrace = new pcap::Trace (traceName.toStdString(), setup, update);
        }
      catch (const pcap::Trace::interrupted&)
	{
	}
      catch (const pcap::Trace::invalid_operation& e)
        {
          QMessageBox::critical(this, tr("Error"), e.what());
        }
      catch (const std::invalid_argument& e)
        {
          QMessageBox::critical(this, tr("Error"), e.what());
        }
      delete dialog;
      if (not pcapTrace)
	return false;

      Trace* const t = new Trace (pcapTrace);
      connect(t, SIGNAL(positionChanged(int)), SLOT(updatePosition(int)));
      connect(t, SIGNAL(positionChanged(int)), t, SLOT(setFocus()));

      tabs_.setCurrentIndex(tabs_.addTab(t, QFileInfo (traceName).fileName()));
      if (tabs_.count() == 1)
	updateTrace();
      return true;
    }

    bool
    MainWindow::openInNewWindow() const
    {
      MainWindow* const w = newWindow(false);

      assert(w);
      if (not w->open())
	{
	  delete w;
	  return false;
	}
      w->show();
      return true;
    }

    MainWindow*
    MainWindow::duplicateWindow() const
    {
      return setupNewWindow(new MainWindow (*this), true);
    }

    MainWindow*
    MainWindow::newWindow(bool show) const
    {
      return setupNewWindow(new MainWindow, show);
    }

    void
    MainWindow::closeTab()
    {
      delete dynamic_cast<Trace*> (trace());
      if (not tabs_.count())
	close();
    };

    void
    MainWindow::filter()
    {
      Trace* const current = trace();
      assert(current);

      const QString	trace_name = current->trace()->file_name().c_str();
      const QString	default_out = trace_name + ".filtered";

      const QString traceName =
	FilterDialog::getFilteredTraceName(this,
					   trace_name,
					   tr("Filter trace"),
					   default_out);

      if (not traceName.isNull())
	openTrace(traceName);
    }

    void
    MainWindow::stepTab(int step)
    {
      const int	count = tabs_.count();

      if (count < 2)
	return;

      // Add an extra "count" to avoid the mathematically unsound
      // behavior of C's modulo (%) operator with negative numbers.
      const int n = (tabs_.currentIndex() + count + step) % count;
      tabs_.setCurrentIndex(n);
    }

    void
    MainWindow::prevTab()
    {
      stepTab(-1);
    }

    void
    MainWindow::nextTab()
    {
      stepTab(1);
    }

    void
    MainWindow::begin()
    {
      assert(trace());

      trace()->setPosition(0);
    };
    void
    MainWindow::end()
    {
      assert(trace());

      trace()->setPosition(trace()->lastPosition());
    };

    void
    MainWindow::forward()
    {
      assert(trace());

      trace()->setPosition(trace()->position() + step_.currentText().toInt());
    };

    void
    MainWindow::backward()
    {
      assert(trace());

      trace()->setPosition(trace()->position() - step_.currentText().toInt());
    };

    void
    MainWindow::about()
    {
      // FIXME: Find a mechanism to auto-magically keep this text in sync
      // FIXME: with README.
      static const char text[] =
	"<h1 align=\"center\" style=\"margin-bottom: 3px\">%1</h1>"
	"<p align=\"center\" style=\"margin-top: 3px; margin-bottom: 22px\">"
	"  <a href=\"%5\">%5</a>"
	"</p>"
	"<p>%1 is a lightweight PCAP file visualizer.</p>"
	"<p>"
	"%1 provides a trace visualizer that is able to work with "
	"huge traces <nobr>(>10 GiB)</nobr>.  Its goals are speed and "
	"low memory requirements.  Despite its design being "
	"protocol-agnostic, it currently handles only Prism and "
	"IEEE 802.11 headers."
	"</p>"
	"<p><dl>"
	"  <dt><b>Version</b></dt>		<dd>%2</dd>"
	"  <dt><b>Report bugs to</b></dt>	<dd><a href=\"%3\">%3</a></dd>"
	"  <dt><b>Main author</b></dt>		<dd>Thomas Claveirole "
	"    &lt;<a href=\"mailto:thomas.claveirole@lip6.fr\">"
	"thomas.claveirole@lip6.fr</a>&gt;</dt>"
	"  <dt><b>Organization</b></dt>		<dd>%4</dd>"
	"</dl></p>"
	"<p style=\"margin-top: 22px;"
	"  font-style: italic\">"
	"  &copy; %4</p>"
	"<p style=\"font-style: italic\">"
	"  This is free software; see the source for copying conditions.  "
	"  There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR "
	"  A PARTICULAR PURPOSE."
	"</p>";

      QMessageBox::about(this,
			 tr("About %1").arg(qApp->applicationName()),
			 tr(text)
			 .arg(qApp->applicationName())		// %1
			 .arg(PACKAGE_VERSION)			// %2
			 .arg(PACKAGE_BUGREPORT)		// %3
			 .arg(qApp->organizationName())		// %4
			 .arg(qApp->organizationDomain()));	// %5
    }

    void
    MainWindow::updateTrace()
    {
      Trace* const t = trace();

      if (t)
	{
	  BOOST_FOREACH(QAction* a, trace_related_actions_)
	    a->setEnabled(true);
	  finder_.setEnabled(true);
	  finder_.setMinimum(1);
	  finder_.setMaximum(t->lastPosition() + 1);

	  const pcap::Trace* const 	pcapT = t->trace();
	  const QString			fileName = pcapT->file_name().c_str();

	  trace_name_.setText(fileName);
	  updatePosition(t->position());
	  trace_size_.setText(tr("%1 MiB")
			      .arg(pcapT->file_size() / float (1 << 20)));

	  setWindowTitle(QFileInfo (fileName).fileName() + " - " +
			 qApp->applicationName());

	  t->setFocus();
	}
      else
	{
	  static const QString	na = tr("N/A");

	  BOOST_FOREACH(QAction* a, trace_related_actions_)
	    a->setEnabled(false);
	  finder_.setEnabled(false);
	  trace_name_.setText(na);
	  position_.setText(na);
	  trace_size_.setText(na);
	  setWindowTitle(qApp->applicationName());
	}

      const bool tabs_actions_enabled = tabs_.count() > 1;
      BOOST_FOREACH(QAction* a, tabs_related_actions_)
	a->setEnabled(tabs_actions_enabled);
    }

    void
    MainWindow::updatePosition(int p)
    {
      assert(trace());

      position_.setText(trace()->positionString());
      finder_.setValue(p + 1);
    }

    void
    MainWindow::FinderEditingFinished()
    {
      assert(trace());

      trace()->setPosition(finder_.value() - 1);
    }

    MainWindow::NewProgressDialog::NewProgressDialog(QProgressDialog**	d,
						     QMainWindow*	w):
      d_ (d),
      w_ (w)
    {
    }

    void
    MainWindow::NewProgressDialog::operator () (std::streampos min,
						std::streampos max)
    {
      QProgressDialog*&	d = *d_;

      d = new QProgressDialog (tr("Building Index...<br />"
				  "<b>Total size:</b> %1 MiB")
			       .arg(max / (1 << 20)),
			       tr("&Cancel"),
			       min >> 16,
			       max >> 16,
			       w_);
      d->setWindowTitle(tr("Building index..."));
      d->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Minimum);
    }

    MainWindow::
    UpdateProgressDialog::UpdateProgressDialog(QProgressDialog** d): d_ (d)
    {
    }

    const std::string*
    MainWindow::UpdateProgressDialog::operator () (std::streampos cur)
    {
      const static std::string	interrupt_message ("User interruption");
      QProgressDialog&		dialog  = **d_;

      dialog.setValue(cur >> 16);
      qApp->processEvents();
      return dialog.wasCanceled() ? &interrupt_message : 0;
    }

    Trace*
    MainWindow::trace()
    {
      QWidget* const	w = tabs_.currentWidget();
      Trace* const	t = dynamic_cast<Trace*> (w);

      assert(not w or t);
      return t;
    }

    void
    MainWindow::initWidgets()
    {
      setWindowIcon(pixmap(tr("wscout")));
      setupMenuBar();
      setupToolBar();
      setupCentralWidget();
      setupStatusBar();
      resize(950, 500);
    }

    void
    MainWindow::setupMenuBar()
    {
      // File.
      {
	QMenu* const	fileMenu = menuBar()->addMenu(tr("&File"));
	QAction* const	o = fileMenu->addAction(tr("&Open..."),
						this, SLOT(open()));
	QAction* const	n = fileMenu->addAction(tr("Open in &new window..."),
						this, SLOT(openInNewWindow()));
	QAction* const	c = fileMenu->addAction(tr("&Close"),
						this, SLOT(closeTab()));
	QAction* const	f = fileMenu->addAction(tr("&Filter..."),
						this, SLOT(filter()));
	fileMenu->addSeparator();
	QAction* const	q = fileMenu->addAction(tr("&Quit"),
						qApp, SLOT(quit()));

	o->setShortcut(QKeySequence::Open);
	n->setShortcut(QKeySequence::New);
	c->setShortcut(QKeySequence::Close);
	f->setShortcut(tr("Ctrl+I"));
	q->setShortcut(tr("Ctrl+Q"));

	trace_related_actions_.push_back(c);
	trace_related_actions_.push_back(f);
      }

      // View.
      {
	QMenu* const	viewMenu = menuBar()->addMenu(tr("&View"));

	QAction* const	w = viewMenu->addAction(tr("New &window"),
						this, SLOT(newWindow()));
	QAction* const	d = viewMenu->addAction(tr("&Duplicate window"),
						this, SLOT(duplicateWindow()));
	viewMenu->addSeparator();


	QAction* const	p = viewMenu->addAction(tr("&Previous tab"),
						this, SLOT(prevTab()));
	QAction* const	n = viewMenu->addAction(tr("&Next tab"),
						this, SLOT(nextTab()));
	viewMenu->addSeparator();

	QAction* const	b = viewMenu->addAction(tr("&Beginning"),
						this, SLOT(begin()));
	QAction* const	a = viewMenu->addAction(tr("Step b&ackward"),
						this, SLOT(backward()));
	QAction* const	f = viewMenu->addAction(tr("Step &forward"),
						this, SLOT(forward()));
	QAction* const	e = viewMenu->addAction(tr("&End"), this, SLOT(end()));

	w->setShortcut(QKeySequence::AddTab);
	d->setShortcut(tr("Ctrl+D"));
	p->setShortcut(QKeySequence::PreviousChild);
	n->setShortcut(QKeySequence::NextChild);
	b->setShortcut(QKeySequence::MoveToStartOfDocument);
	a->setShortcut(QKeySequence::Back);
	f->setShortcut(QKeySequence::Forward);
	e->setShortcut(QKeySequence::MoveToEndOfDocument);

	tabs_related_actions_.push_back(p);
	tabs_related_actions_.push_back(n);

	trace_related_actions_.push_back(b);
	trace_related_actions_.push_back(a);
	trace_related_actions_.push_back(f);
	trace_related_actions_.push_back(e);
      }

      // Help.
      {
	QMenu* const	helpMenu = menuBar()->addMenu(tr("&Help"));
	helpMenu->addAction(tr("About &%1...").arg(qApp->applicationName()),
			    this, SLOT(about()));
	helpMenu->addAction(tr("About &Qt..."), qApp, SLOT(aboutQt()));
      }
    }

    void
    MainWindow::setupToolBar()
    {
      QToolBar* const t = new QToolBar;

      // Trace operations.
      {
	t->addAction(pixmap(tr("fileopen")),
		     tr("Open..."), this, SLOT(open()));
	QAction* const c = t->addAction(pixmap(tr("fileclose")),
					tr("Close"), this, SLOT(closeTab()));
	QAction* const f = t->addAction(pixmap(tr("filter")),
					tr("Filter..."), this, SLOT(filter()));

	trace_related_actions_.push_back(c);
	trace_related_actions_.push_back(f);
      }

      // Window operations.
      {
	t->addSeparator();
	t->addAction(pixmap(tr("window_new")), tr("New window"),
		     this, SLOT(newWindow()));
	t->addAction(pixmap(tr("window_dup")), tr("Duplicate window"),
		     this, SLOT(duplicateWindow()));
	t->addAction(pixmap(tr("window_openfile")), tr("Open in new window..."),
		     this, SLOT(openInNewWindow()));
      }

      // Step chooser.
      {
	step_.setEditable(true);
	step_.setInsertPolicy(QComboBox::InsertAlphabetically);
	step_.setToolTip("Step amount");
	// FIXME: Memory leak!
	step_.setValidator(new QIntValidator (1, int (1e9), &step_));
	step_.addItem("1");
	step_.addItem("5");
	step_.setCurrentIndex(1);
	step_.addItem("10");
	step_.addItem("25");
	step_.addItem("50");
	step_.addItem("100");
	step_.addItem("500");
	step_.addItem("1000");
	step_.addItem("5000");
	step_.addItem("10000");
	step_.addItem("100000");
	t->addWidget(&step_);
      }

      // Navigation misc.
      {
	QAction* const b = t->addAction(pixmap(tr("player_start")),
					tr("Beginning"),
					this, SLOT(begin()));
	QAction* const a = t->addAction(pixmap(tr("player_rew")),
					tr("Step backward"),
					this, SLOT(backward()));
	QAction* const f = t->addAction(pixmap(tr("player_fwd")),
					tr("Step forward"),
					this, SLOT(forward()));
	QAction* const e = t->addAction(pixmap(tr("player_end")),
					tr("End"), this, SLOT(end()));

	trace_related_actions_.push_back(b);
	trace_related_actions_.push_back(a);
	trace_related_actions_.push_back(f);
	trace_related_actions_.push_back(e);
      }

      // Finder.
      {
	finder_.setPrefix("#");
	finder_.setSizePolicy(QSizePolicy (QSizePolicy::Expanding,
					   QSizePolicy::Fixed));
	finder_.setToolTip("Goto packet");
	connect(&finder_,
		SIGNAL(editingFinished()),
		SLOT(FinderEditingFinished()));
	connect(&focus_finder_,	SIGNAL(activated()),
		&finder_,	SLOT(clear()));
	connect(&focus_finder_,	SIGNAL(activated()),
		&finder_,	SLOT(setFocus()));
	t->addWidget(&finder_);
      }
      addToolBar(t);
    }

    void
    MainWindow::setupStatusBar()
    {
      const QSizePolicy	p (QSizePolicy::MinimumExpanding,
			   QSizePolicy::Minimum);
      QStatusBar* const s = statusBar();

      trace_name_.setSizePolicy(p);
      position_.setSizePolicy(p);
      trace_size_.setSizePolicy(p);

      trace_name_.setContentsMargins(status_bar_hmargin, status_bar_vmargin,
				     status_bar_hmargin, status_bar_vmargin);
      position_.setContentsMargins(status_bar_hmargin, status_bar_vmargin,
				   status_bar_hmargin, status_bar_vmargin);
      trace_size_.setContentsMargins(status_bar_hmargin, status_bar_vmargin,
				     status_bar_hmargin, status_bar_vmargin);

      s->addWidget(&trace_name_);
      s->addWidget(&position_);
      s->addWidget(&trace_size_);
    }

    void
    MainWindow::setupCentralWidget()
    {
      setCentralWidget(&tabs_);
      connect(&tabs_, SIGNAL(currentChanged(int)), SLOT(updateTrace()));
    }

    MainWindow*
    MainWindow::setupNewWindow(MainWindow* w, bool show)
    {
      w->setAttribute(Qt::WA_DeleteOnClose);
      if (show)
	w->show();
      return w;
    }

  } // End of namespace gui.

} // End of namespace wscout.
