/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <QtCore/QProcess>
#include <QtCore/QDir>
#include <QtGui/QImageReader>

#include <boost/foreach.hpp>

#include "wscout_config.hh"
#include "wscout_gui_pixmap.hh"

namespace wscout
{

  namespace gui
  {

    namespace
    {
      static const QString	pixmap_search_path_prefix = "pixmaps";
    }

    /**
     * Initialize the pixmap search paths.
     *
     * The pixmap search paths are composed of:
     *
     * <ol>
     *
     *   <li>The directories listed in the @c PIXMAP_PATH environment variable,
     *       separated by semi-columns,</li>
     *
     *   <li> @code $(top_builddir)/pixmaps @endcode </li>
     *
     *   <li> @code $(top_srcdir)/pixmaps @endcode </li>
     *
     *   <li> @code $(pixmapdir) @endcode
     *        (@c /usr/local/share/wscout/pixmaps by default) </li>
     *
     *   <li>The application's internal resources, in the @c /pixmaps
     *       directory.</li>
     *
     * </ol>
     *
     * Pixmaps are searched in those directories in ascending order.
     */
    void	pixmapInit()
    {
      const QStringList	env = QProcess::systemEnvironment();

      BOOST_FOREACH(const QString& s, env)
	{
	  const int e = s.indexOf('=');
	  if (e < 1)
	    continue;

	  const QString	name = s.left(e);
	  const QString	value = s.right(s.size() - e - 1);

	  if (name == WSCOUT_PIXMAP_PATH_ENVVAR_NAME)
	    {
	      QDir::setSearchPaths(pixmap_search_path_prefix,
				   value.split(':', QString::SkipEmptyParts));
	      break;
	    }
	}
      QDir::addSearchPath(pixmap_search_path_prefix,
			  WSCOUT_TOP_BUILDDIR "/pixmaps");
      QDir::addSearchPath(pixmap_search_path_prefix,
			  WSCOUT_TOP_SRCDIR "/pixmaps");
      QDir::addSearchPath(pixmap_search_path_prefix, WSCOUT_PIXMAP_DIR);
      QDir::addSearchPath(pixmap_search_path_prefix, ":/pixmaps");
    }

    /**
     * Search and retrieve a pixmap inside the pixmap search paths.
     *
     * @e e.g. when given @c "foo" this function might search for
     * @c /some/path/foo.svg, @c /some/other/path/foo.png, etc.
     */
    QPixmap	pixmap(const QString& name)
    {
      QList<QByteArray> supported = QImageReader::supportedImageFormats();

      BOOST_FOREACH(const QByteArray& f, supported)
	{
	  QPixmap r (pixmap_search_path_prefix + ':' + name + '.' + f);

	  if (not r.isNull())
	    return r;
	}
      return QPixmap ();
    }

  } // End of namespace gui.

} // End of namespace wscout.
