/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include "wscout_pcap_trace.hh"

namespace wscout
{

  namespace pcap
  {

    namespace
    {

      static
      void
      delete_uint8_cxx_array(uint8_t* const p)
      {
	delete[] p;
      }

    } // End of anonymous namespace.

    Trace::PacketIterator::Packet::Packet(std::ifstream& fstream,
					  bool swap,
					  int32_t zone)
    {
      pcap_packet_header	h;

      if (not fstream.read(reinterpret_cast<char*> (&h), sizeof (h)) or
	  fstream.gcount() != sizeof (h))
	throw invalid_operation ("Unreadable packet header");
      if (swap)
	{
	  ts.tv_sec = swap_s32(h.ts.tv_sec);
	  ts.tv_usec = swap_s32(h.ts.tv_usec);
	  caplen = swap_u32(h.caplen);
	  len = swap_u32(h.len);
	}
      else
	{
	  ts.tv_sec = h.ts.tv_sec;
	  ts.tv_usec = h.ts.tv_usec;
	  caplen = h.caplen;
	  len = h.len;
	}
      ts.tv_sec += zone;
      data.reset(new uint8_t[caplen], delete_uint8_cxx_array);
      if (not fstream.read(reinterpret_cast<char*> (data.get()), caplen) or
	  fstream.gcount() != caplen)
	throw invalid_operation ("Unreadable packet data");
    }

    bool
    Trace::PacketIterator::Packet::operator == (const Packet& rhs) const
    {
      return (ts.tv_sec == rhs.ts.tv_sec and ts.tv_usec == rhs.ts.tv_usec and
	      caplen == rhs.caplen and
	      len == rhs.len and
	      0 == memcmp(data.get(), rhs.data.get(), caplen));
    }

    bool
    Trace::PacketIterator::Packet::operator != (const Packet& rhs) const
    {
      return not (*this == rhs);
    }

    Trace::PacketIterator::PacketIterator(Trace& t, bool end_of_trace):
      v_ (end_of_trace ?
	  optional_value (boost::none_t ()) :
	  optional_value (Packet (t.file_, t.swap_, t.zone_))),
      t_ (&t)
    {
    }

    bool
    Trace::PacketIterator::operator == (const PacketIterator& rhs) const
    {
      return t_ == rhs.t_ and v_ == rhs.v_;
    }

    bool
    Trace::PacketIterator::operator != (const PacketIterator& rhs) const
    {
      return not (*this == rhs);
    }

    const Trace::PacketIterator::Packet&
    Trace::PacketIterator::operator * () const
    {
      return v_.get();
    }

    const Trace::PacketIterator::Packet*
    Trace::PacketIterator::operator -> () const
    {
      return v_.get_ptr();
    }

    Trace::PacketIterator&
    Trace::PacketIterator::operator ++ ()
    {
      try
	{
	  v_ = Packet (t_->file_, t_->swap_, t_->zone_);
	}
      catch (const invalid_operation&)
	{
	  v_ = boost::none_t ();
	  t_->file_.clear();
	}
      return *this;
    }

    Trace::PacketIterator
    Trace::PacketIterator::operator ++ (int)
    {
      PacketIterator r (*this);

      ++(*this);
      return r;
    }

    Trace::PacketIterator
    Trace::operator [] (size_t i)
    {
      size_t	near_mark = i / mark_step;
      size_t	extra_steps = i % mark_step;

      assert(i < pkt_count_);
      assert(near_mark < marks_.size());

      if (not file_.seekg(marks_[near_mark], std::ios::beg))
	throw invalid_operation ("Seek operation failed");

      PacketIterator	r (*this, false);
      for (unsigned j = 0; j < extra_steps; ++j)
	{
	  ++r;
	  assert(r != end_of_trace());
	}
      return r;
    }

    Trace::PacketIterator
    Trace::end_of_trace()
    {
      return PacketIterator (*this, true);
    }

    size_t
    Trace::size() const
    {
      return pkt_count_;
    }

    Trace::link_type
    Trace::linktype() const
    {
      return type_;
    }

    std::streampos
    Trace::file_size() const
    {
      return file_size_;
    }

    const std::string&
    Trace::file_name() const
    {
      return file_name_;
    }

    void
    Trace::warn(const std::streampos& p, const std::string& msg)
    {
      std::cerr << file_name_ <<  ':' << p << ": WARNING: " << msg << '.'
		<< std::endl;
    }

    void
    Trace::handle_truncation(const std::string& msg)
    {
      warn(file_size_, "corrupted file (" + msg + ')');
      if (not pkt_count_)
	throw std::invalid_argument ("Empty PCAP file");
      --pkt_count_;
    }

  } // End of namespace pcap.

} // End of namespace wscout.
