/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#ifndef WSCOUT_PCAP_TRACE_HH_
# define WSCOUT_PCAP_TRACE_HH_

extern "C"
{
# include <sys/types.h>
# include <sys/times.h>
# include <inttypes.h>
}

# include <fstream>
# include <cstdlib>
# include <iterator>
# include <vector>
# include <stdexcept>

# include <boost/optional.hpp>
# include <boost/shared_ptr.hpp>

namespace wscout
{

  /// PCAP file loading.
  namespace pcap
  {

    /// PCAP trace file designed for fast random access.
    struct Trace
    {

      /// Iterator over PCAP packets.
      struct PacketIterator
      {

	/// PCAP packet.
	struct Packet
	{
	  struct timeval		ts;	///< Time of arrival.
	  uint32_t			caplen;	///< Number of captured bytes.
	  uint32_t			len;	///< %Packet len.
	  boost::shared_ptr<uint8_t>	data;	///< %Packet data.

	  /// Construct a PCAP packet from a file.
	  Packet(std::ifstream& fstream, bool swap, int32_t zone);

	  /// @name Equality comparison.
	  //@{
	  bool operator == (const Packet& rhs) const;
	  bool operator != (const Packet& lhs) const;
	  //@}

	}; // End of struct Packet.

	/// @name Iterator types.
	//@{
	typedef ssize_t			difference_type;
	typedef Packet			value_type;
	typedef value_type*		pointer_type;
	typedef value_type&		reference_type;
	typedef std::input_iterator_tag	iterator_category;
	//@}

	/**
	 * Build an iterator over PCAP packets from a PCAP trace.
	 *
	 * If @c end_of_trace, then build an iterator that
	 * represents the trace's end position.
	 */
	PacketIterator(Trace& t, bool end_of_trace = true);

	/// @name Equality comparison.
	//@{
	bool		operator == (const PacketIterator& rhs) const;
	bool		operator != (const PacketIterator& rhs) const;
	//@}

	/// @name Value access.
	//@{
	const Packet&	operator * () const;
	const Packet*	operator -> () const;
	//@}

	/// @name Increment.
	//@{
	PacketIterator&	operator ++ ();
	PacketIterator	operator ++ (int);
	//@}

      private:
	typedef boost::optional<value_type>	optional_value;

	optional_value	v_;	///< If not an end iterator, its value.
	Trace*		t_;	///< Corresponding trace.
      };

      /// A file operation (open, seek, etc.) reported an error.
      struct invalid_operation: public std::runtime_error
      {
	/// Construct the exception given a description.
	invalid_operation(const std::string&);
      };

      /// A callback asked the trace construction to stop.
      struct interrupted: public std::runtime_error
      {
	/// Construct the exception given a description.
	interrupted(const std::string&);
      };

      /// PCAP packet iterator.
      typedef PacketIterator	iterator;

      /// PCAP file link types.
      enum link_type
	{
	  LOOP		= 0,	///< BSD loopback encapsulation
	  EN10MB	= 1,	///< Ethernet (10Mb)
	  EN3MB		= 2,	///< Experimental Ethernet (3Mb)
	  AX25		= 3,	///< Amateur Radio AX.25
	  PRONET	= 4,	///< Proteon ProNET Token Ring
	  CHAOS		= 5,	///< Chaos
	  IEEE802	= 6,	///< IEEE 802 Networks
	  ARCNET	= 7,	///< ARCNET, with BSD-style header
	  SLIP		= 8,	///< Serial Line IP
	  PPP		= 9,	///< Point-to-point Protocol
	  FDDI		= 10,	///< FDDI
	  IEEE802_11	= 105,	///< IEEE 802.11 wireless
	  PRISM_HEADER	= 119	///< Prism PHY-layer information
	};

      /// Construct a trace from a file name.
      Trace(const std::string&);

      /**
       * Construct a trace from a file name, with progress indicator
       * callbacks.
       *
       * @param[in]	filename	The file's name.
       * @param[in]	pi_setup	Called before reading packets with
       *				two arguments.  The first one is
       *				the first packet's offset, the last
       *				one is the file size.
       * @param[in]	pi_update	Called periodically when reading
       *				packets.  The current offset in the
       *				file is passed as argument.  Returns
       *				either 0 (continue) or a pointer to
       *				a string (stop and throws
       *				@c interrupted(string) ).
       */
      template <typename ProgressInfoSetupType,
		typename ProgressInfoUpdateType>
      Trace(const std::string&		filename,
	    ProgressInfoSetupType&	pi_setup,
	    ProgressInfoUpdateType&	pi_update);

      /// Get an iterator pointing to a specific packet.
      PacketIterator		operator [] (size_t);

      /// Get an iterator to the end of trace.
      PacketIterator		end_of_trace();

      /// Get the number of packets in the trace.
      size_t			size() const;

      /// Get the trace's link type.
      link_type			linktype() const;

      /// Get the trace size in bytes.
      std::streampos		file_size() const;

      /// Get the trace's name.
      const std::string&	file_name() const;

    protected:

      enum
	{
	  /**
	   * Register a mark every @c mark_step packets.
	   *
	   * A mark is an offset in the trace file.  Marks are
	   * stored in a vector at regular intervals, so fast
	   * access to random packets is possible.
	   *
	   * This is a trade-off: more marks allows faster access, but
	   * fewer marks uses less memory.
	   */
	  mark_step	= 512,

	  /**
	   * When loading traces, read packet data with steps of @c
	   * junk_len bytes.
	   *
	   * The more the better, but one must be careful not to grow
	   * the stack too much.  We use reads here instead of seeks
	   * not to break the underlying C++ buffering.  Tests showed
	   * this is faster.
	   */
	  junk_len	= 4096
	};

      /// PCAP file format internal header.
      struct pcap_file_header
      {
	uint32_t	magic;		///< File magic.
	uint16_t	version_major;	///< PCAP format version, major.
	uint16_t	version_minor;	///< PCAP format version, minor.
	int32_t		thiszone;	///< Time offset.
	int32_t		sigfigs;	///< Timestamp accuracy.
	int32_t		snaplen;	///< Max. number of captured bytes.
	int32_t		linktype;	///< PCAP link type.
      };

      /// PCAP file format representation of a packet header.
      struct pcap_packet_header
      {
	struct timeval	ts;	///< Timestamp.
	uint32_t	caplen;	///< Number of captured bytes.
	uint32_t	len;	///< %Packet len.
      };

    private:
      template <typename ProgressInfoSetupType,
		typename ProgressInfoUpdateType>
      void
      setup(const std::string&			filename,
	    ProgressInfoSetupType&		pi_setup,
	    ProgressInfoUpdateType&		pi_update);

      void		warn(const std::streampos&p, const std::string& msg);
      void		handle_truncation(const std::string& msg);

      template <typename ProgressInfoSetupType,
		typename ProgressInfoUpdateType>
      void
      setup_marks(ProgressInfoSetupType&	pi_setup,
		  ProgressInfoUpdateType&	pi_update);

      std::ifstream			file_;
      std::vector<std::streampos>	marks_;

      bool				swap_;
      int32_t				zone_;
      link_type				type_;
      size_t				pkt_count_;
      std::streampos			file_size_;
      std::string			file_name_;
    };

  } // End of namespace pcap.

} // End of namespace wscout.

# include "wscout_pcap_trace.hxx"

#endif // ! WSCOUT_PCAP_TRACE_HH_
