/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#ifndef WSCOUT_PCAP_TRACE_HXX_
# define WSCOUT_PCAP_TRACE_HXX_

# include <iostream>

# ifdef TIME_FILE_LOADING
#  include <boost/date_time/posix_time/posix_time.hpp>
# endif // BENCH_FILE_LOADING

# include "wscout_pcap_trace.hh"

namespace wscout
{

  namespace pcap
  {

    /// Change endianness a unsigned 32 bit word.
    inline
    uint32_t
    swap_u32(uint32_t i)
    {
      return (((i & 0xFF)	<< 24) |
	      ((i & 0xFF00)	<<  8) |
	      ((i & 0xFF0000)	>>  8) |
	      (i >> 24));
    }

    /// Change endianness a signed 32 bit word.
    inline
    int32_t
    swap_s32(int32_t i)
    {
      return int32_t (swap_u32(uint32_t (i)));
    }

    namespace
    {

      struct DummySetup
      {
	inline
	void
	operator () (std::streampos, std::streampos)
	{
	}
      };

      struct DummyUpdate
      {
	inline
	const std::string*
	operator () (std::streampos)
	{
	  return 0;
	}
      };

    } // End of anonymous namespace.

    inline
    Trace::invalid_operation::invalid_operation(const std::string& s):
      std::runtime_error (s)
    {
    }

    inline
    Trace::interrupted::interrupted(const std::string& s):
      std::runtime_error (s)
    {
    }

    inline
    Trace::Trace(const std::string& filename)
    {
      DummySetup	pi_setup;
      DummyUpdate	pi_update;

      setup(filename, pi_setup, pi_update);
    }

    template <typename ProgressInfoSetupType,
	      typename ProgressInfoUpdateType>
    Trace::Trace(const std::string&		filename,
		 ProgressInfoSetupType&		pi_setup,
		 ProgressInfoUpdateType&	pi_update)
    {
      setup(filename, pi_setup, pi_update);
    }

    template <typename ProgressInfoSetupType,
	      typename ProgressInfoUpdateType>
    void
    Trace::setup(const std::string&		filename,
		 ProgressInfoSetupType&		pi_setup,
		 ProgressInfoUpdateType&	pi_update)
    {
      pcap_file_header	h;

      file_.open(filename.c_str());
      if (not file_)
	throw invalid_operation ("Open failed");
      file_name_ = filename;
      if (not file_.seekg(0, std::ios::end))
	throw invalid_operation ("Seek operation failed");
      file_size_ = file_.tellg();
      if (not file_.seekg(0, std::ios::beg))
	throw invalid_operation ("Seek operation failed");

      if (not file_.read(reinterpret_cast<char*> (&h), sizeof (h)) or
	  file_.gcount() != sizeof (h))
	throw invalid_operation ("Unreadable PCAP file header");
      switch (h.magic)
	{
	case 0xa1b2c3d4:
	  swap_ = false;
	  break;
	case 0xd4c3b2a1:
	  swap_ = true;
	  break;
	default:
	  throw std::invalid_argument ("Not a PCAP trace");
	}
      zone_ = swap_ ? swap_s32(h.thiszone) : h.thiszone;
      type_ = link_type (swap_ ? swap_s32(h.linktype) : h.linktype);

      setup_marks(pi_setup, pi_update);
    }

    template <typename ProgressInfoSetupType,
	      typename ProgressInfoUpdateType>
    void
    Trace::setup_marks(ProgressInfoSetupType&	pi_setup,
		       ProgressInfoUpdateType&	pi_update)
    {
# ifdef TIME_FILE_LOADING
      using namespace boost::posix_time;

      const ptime start = microsec_clock::universal_time();
# endif // TIME_FILE_LOADING

      pkt_count_ = 0;

      {
	const std::streampos	first_packet_offset = file_.tellg();

	marks_.push_back(first_packet_offset);
	pi_setup(first_packet_offset, file_size_);
      }

      pcap_packet_header	h;

      while (file_.read(reinterpret_cast<char*> (&h), sizeof (h)))
	{
	  if (file_.gcount() != sizeof (h))
	    {
	      handle_truncation("truncated packet header");
	      break;
	    }

	  char			junk[junk_len];

	  if (swap_)
	    h.caplen = swap_u32(h.caplen);
	  while(h.caplen) // Skip packet data.
	    {
	      size_t		to_read = std::min(sizeof (junk),
						   size_t (h.caplen));

	      if (not file_.read(junk, to_read))
		{
		  handle_truncation("truncated packet data");
		  break;
		}
	      h.caplen -= to_read;
	    }

	  if (++pkt_count_ % mark_step == 0)
	    {
	      const std::streampos	current_offset = file_.tellg();

	      marks_.push_back(current_offset);
	      if (const std::string* e = pi_update(current_offset))
		throw interrupted (*e);
	    }
	}
      pi_update(file_size_);
      if (file_.bad())
	throw invalid_operation ("Unrecoverable corruption in PCAP file");
      file_.clear();

# ifdef TIME_FILE_LOADING
      std::cerr << (microsec_clock::universal_time() - start) << std::endl;
# endif // TIME_FILE_LOADING
    }

  } // End of namespace pcap.

} // End of namespace wscout.

#endif // ! WSCOUT_PCAP_TRACE_HXX_
