/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_DESCRIPTOR_HH_
# define PCAP_DESCRIPTOR_HH_

# include <string>

# include <boost/optional.hpp>
# include <boost/shared_ptr.hpp>

# include <trace-tools/tool/iterator_base.hh>
# include <trace-tools/tool/iterable.hh>
# include <trace-tools/pcap/frame_iterator.hh>
# include <trace-tools/pcap/typedefs.hh>
# include <trace-tools/pcap/dumper.hh>

/// PCAP and @c libpcap related compounds.
namespace pcapxx
{

  template <class D>
  struct descriptor;

} // End of namespace pcapxx.

namespace tool
{

  /// Virtual %types for pcapxx::descriptor.  @see types
  template <class B>
  struct types< pcapxx::descriptor<B> >
  {
    /// @link pcapxx::descriptor PCAP descriptor @endlink iterator type.
    typedef pcapxx::internals::frame_iterator<B, bottom>	iterator;
  };

} // End of namespace tool.

namespace pcapxx
{

  /**
   * Enable various PCAP operations.
   *
   * A PCAP %descriptor allows you to read frames (or packets) either
   * from PCAP files or network interfaces.  You may also use a PCAP
   * %descriptor to retrieve a dumper object, in order to dump frames
   * (or packets) into a PCAP file.
   *
   * A PCAP %descriptor is an iterable object.  You may iterate over
   * frames like this:
   *
   * @code
   * pcapxx::descriptor<> d ("file.pcap");
   *
   * for (pcapxx::descriptor<>::iterator i = d.begin(); i != d.end(); ++i)
   *   {
   *     // Access to i->pcap_header and i->bytes.
   *     //
   *     // *i is of type frame_descriptor.
   *   }
   * @endcode
   *
   * @see dumper, frame_descriptor, internals::frame_iterator
   */
  template <class Bottom = tool::bottom>
  struct descriptor: tt_inherit(public tool::iterable, descriptor<Bottom>)
  {
    /**
     * Construct a descriptor for live packet capture.
     *
     * @param[in]  device	Network device's name.
     * @param[in]  snaplen	Maximum number of bytes to capture per frame.
     * @param[in]  promisc	Put the interface in promiscuous mode if true.
     * @param[in]  to_ms	Read timeout, in milliseconds.
     */
    descriptor(const std::string& device, int snaplen, int promisc, int to_ms);

    /// Construct a descriptor for offline file access.
    descriptor(const std::string& fname);

    /**
     * Construct a descriptor not for packet capture.
     *
     * @param[in]  linktype	Frames' link type.
     * @param[in]  snaplen	Maximum number of bytes to dump per frame.
     */
    descriptor(int linktype, int snaplen);

    /**
     * Frames' link type.
     *
     * @see @c libpcap documentation about link %types.
     */
    int			datalink() const;

    /// File endianness @e w.r.t this machine.
    bool		is_swapped() const;

    /// Last error message.
    std::string		error() const;

    /// Get a dumper object for a given file.
    pcapxx::dumper	dumper(const std::string& fname) const;

    /**
     * Read frames.
     *
     * @param[in]  Callback	Function to call on each frame.
     * @param[in]  user		User data for @c callback.
     * @param[in]  cnt		Maximum number of frames to read.
     *				-1 is unlimited.
     */
    template <handler Callback>
    int
    loop(void* user = 0, int cnt = -1);

  private:

    /// User data for @c handler_wrapper().
    struct wrapper_data
    {
      /// Points to the descriptor that triggered the callback call.
      descriptor*	desc;

      /// The actual user data to pass to the user callback.
      void*		user;
    };

    /// Wraps callback calls from pcap_loop.
    template <handler H>
    static
    void
    handler_wrapper(u_char*, const pcap_pkthdr*, const u_char*);

    /// Data shared by identical PCAP descriptors.
    struct shared_data
    {
      shared_data(pcap_t*);
      ~shared_data();

      pcap_t*	desc;	///< Actual @c libpcap PCAP %descriptor.
      unsigned	curr;	///< Frame count - updated in handler_wrapper().
    };

    boost::shared_ptr<shared_data> d_;
  };

} // End of namespace pcapxx.

# include "descriptor.hxx"

#endif // ! PCAP_DESCRITPOR_HH_
