/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_FRAME_DESCRIPTOR_HH_
# define PCAP_FRAME_DESCRIPTOR_HH_

# include <ostream>
# include <boost/shared_ptr.hpp>

# include <trace-tools/pcap/pkthdr.hh>

namespace pcapxx
{

  /**
   * Describe a PCAP frame/packet.
   *
   * A frame %descriptor holds a frame's @link pkthdr PCAP header
   * @endlink and bytes.
   *
   * Upon construction, both the frame's PCAP header and bytes are
   * deep-copied into the %descriptor.  You may therefore work with the
   * frame %descriptor after the original frame information has been
   * invalidated.
   *
   * Frame descriptors use shared pointers to store frame information.
   * You may copy them efficiently.  You may also retrieve smart pointer
   * to both the frame's PCAP header and bytes.
   */
  struct frame_descriptor
  {
    /// Smart pointer to a frame's PCAP header.
    typedef boost::shared_ptr<const pkthdr>	pkthdr_ptr;

    /// Smart pointer to a frame's bytes.
    typedef boost::shared_ptr<const uint8_t>	bytes_ptr;

    /**
     * Build a PCAP frame %descriptor from corresponding @link pkthdr
     * PCAP header @endlink and bytes.
     *
     * The content of @c header and @c bytes are deep-copied into the
     * frame %descriptor at construction.  The %descriptor does
     * therefore remain valid even after @c header and @c bytes have
     * bee invalidated.
     */
    frame_descriptor(const pkthdr* header, const void* bytes);

    /// Get a smart pointer to the frame's PCAP header.
    const pkthdr_ptr&	pcap_header() const;

    /// Get a smart pointer to the frame's data.
    const bytes_ptr&	bytes() const;

    /// Dump a frame %descriptor to a C++ stream.
    std::ostream&	print(std::ostream&) const;

    /// @name Equality checks
    //@{
    bool	operator == (const frame_descriptor&) const;
    bool	operator != (const frame_descriptor&) const;
    //@}

  private:
    static void	delete_uint8_t_array(uint8_t*);

    pkthdr_ptr	pcap_header_;
    bytes_ptr	bytes_;
  };

  /// Dump a @link frame_descriptor frame %descriptor @endlink to a C++ stream.
  std::ostream&
  operator << (std::ostream&, const frame_descriptor&);

} // End of namespace pcapxx.

# include "frame_descriptor.hxx"

#endif // ! PCAP_FRAME_DESCRITPOR_HH_
