/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_ITERABLE_HH_
# define TOOL_ITERABLE_HH_

# include <trace-tools/tool/static_base.hh>

namespace tool
{

  template <class B>
  struct iterable;

  /// Virtual %types for iterable. @see types
  template <class B>
  struct types< iterable<B> >
  {
    typedef undefined_type	iterator; /// Iterator for %iterable objects.
  };

  /**
   * Base class to represent %iterable objects.
   *
   * An %iterable oject aggregates elements on which iteration is
   * possible.  iterable provides basic mechanisms to permorm this
   * iteration.  That is, it provides an iterable::iterator type,
   * begin() / end() methods, and operator()().
   *
   * In order to subclass iterable, you must define an %iterator type
   * for your %iterable class.  This type must have a similar
   * constructor:
   *
   * @code
   * my_iterator::my_iterator(my_iterable& e, bool at_end);
   * @endcode
   *
   * The first argument refers to the object you want to iterate on.
   * The second one indicates whether to point the iterator at the
   * begining or past the end of this object's elements.
   *
   * Once you defined this %iterator type, just declare it as part of
   * your %iterable class' @link types virtual types @endlink:
   *
   * @code
   * template <MyIterableParams>
   * struct types< MyIterable<MyIterableParams> >
   * {
   *   typedef my_iterator iterator;
   * };
   * @endcode
   *
   * @see iterator
   */
  template <class Bottom = bottom>
  struct iterable: tt_inherit(public static_base, iterable<Bottom>)
  {
    /// Exact type for this template instantiation.
    typedef tt_get_exact(Bottom, iterable<Bottom>)	exact_type;

    /// Iterator type for this template instantiation.
    typedef tt_type(iterator, exact_type)		iterator;

    /// Get an iterator to the first element.
    iterator	begin();

    /// Get an iterator to the end (past the last element).
    iterator	end();

    /// Copy elements to @c output .
    template <class OutputIterator>
    void
    operator () (const OutputIterator& output);
  };

} // End of namespace tool.

# include "iterable.hxx"

#endif // ! TOOL_ITERABLE_HH_
