/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_ITERATOR_BASE_HH_
# define TOOL_ITERATOR_BASE_HH_

# include <boost/optional.hpp>

# include <trace-tools/tool/iterator.hh>

namespace tool
{

  template <class B>
  struct iterator_base;

  /// Virtual %types for iterator_base. @see types
  template <class B>
  struct types< iterator_base<B> >
  {
    /// The iterator's value type.
    typedef undefined_type	value_type;

    /// Type of the iterable structure the iterator is obtained from.
    typedef undefined_type	iterable_type;
  };

  /**
   * Common implementation for most iterators.
   *
   * iterator_base implements equal(), get(), and get_ptr() from
   * iterator.  iterator_base embeds its value inside a @c
   * boost::optional<value_type> attribute.  To subclass it, just
   * implement the @c increment() method and define the value_type and
   * iterable_type @link types virtual types@endlink.  @c increment()
   * shall retrieve the %iterator's next value and store it into @c v_
   * .  Store @c boost::none_t() into @c v_ to indicate the end.
   *
   * You may access the %iterator's iterable object through the @c i_
   * attribute.
   */
  template <class Bottom>
  struct iterator_base: tt_inherit(public iterator, iterator_base<Bottom>)
  {
    /// Exact type for this template instantiation.
    typedef tt_get_exact(Bottom, iterator_base<Bottom>)	exact_type;

    /// This iterator's value type.
    typedef tt_type(value_type, exact_type)		value_type;

    /// Type of the iterable structure the iterator is obtained from.
    typedef tt_type(iterable_type, exact_type)		iterable_type;

    /// @link iterable @endlink -compatible constructor.
    iterator_base(iterable_type& i, bool end);

    /// Equality check. @see iterator
    bool		equal(const iterator_base&) const;

    /// Indirection as a reference.  @see iterator
    const value_type&	get() const;

    /// Indirection as a pointer.  @see iterator
    const value_type*	get_ptr() const;

  protected:
    /// Protected destructor: iterator_base is abstract.
    ~iterator_base();

  protected:
    /// Type for @link v_ @endlink.
    typedef typename boost::optional<value_type>	optional_value;

    /// Pointer to the %iterable object.
    iterable_type*	i_;

    /// Current value, or @c boost::none_t.
    optional_value	v_;
  };

} // End of namespace tool.

# include "iterator_base.hxx"

#endif // ! TOOL_ITERATOR_BASE_HH_
