/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_SI_HH_
# define TOOL_SI_HH_

# include <cstdlib>
# include <string>
# include <ostream>

namespace tool
{

  /// Internal compounds - Not designed for users.
  namespace internals
  {

    /**
     * Helper class to display human readable bit sizes.
     *
     * @e E.g. @code
     * typedef SI_unit<8>	byte;
     * typedef SI_unit<8192>	kilobyte;
     *
     * const int i = 65536;
     *
     * std::cout << i << " bits is "
     *           << byte (i) << " and "
     *           << kilobyte (i) << std::endl;
     * @endcode
     */
    template <size_t OneUnitInBits>
    struct SI_unit
    {
      /// Construct a SI_unit from a number of bits.
      SI_unit(size_t);

      /// Automatic conversion from one SI_unit to another.
      template <size_t OneRHSUnitInBits>
      SI_unit(const SI_unit<OneRHSUnitInBits>&);

      /// Retrieve the stored number of bits.
      size_t			bits() const;

      /// Retrieve the stored number of bytes.
      size_t			bytes() const;

      /// Get this unit's abbreviation.
      static std::string	abbrev();

    private:
      size_t	bits_;

      /// Unit abbreviation (@e e.g. "Mb" for megabits).
      static const char*	abbrev_;
    };

    /// Display a SI_unit object.
    template <size_t OneUnitInBits>
    std::ostream&
    operator << (std::ostream&, const SI_unit<OneUnitInBits>&);

  } // End of namespace internals.

  /// Bits. @see internals::SI_unit
  typedef internals::SI_unit<1>		b;

  /// Kilobits. @see internals::SI_unit
  typedef internals::SI_unit<1000>	kb;

  /// Kibibits. @see internals::SI_unit
  typedef internals::SI_unit<1024>	Kib;

  /// Megabits. @see internals::SI_unit
  typedef internals::SI_unit<1000000>	Mb;

  /// Mebibits. @see internals::SI_unit
  typedef internals::SI_unit<1048576>	Mib;


  /// Bytes. @see internals::SI_unit
  typedef internals::SI_unit<8>		B;

  /// Kilobytes. @see internals::SI_unit
  typedef internals::SI_unit<8000>	kB;

  /// Kibibytes. @see internals::SI_unit
  typedef internals::SI_unit<8192>	KiB;

  /// Megabytes. @see internals::SI_unit
  typedef internals::SI_unit<8000000>	MB;

  /// Mebibytes. @see internals::SI_unit
  typedef internals::SI_unit<8388608>	MiB;

} // End of namespace tool.

# include "si.hxx"

#endif // ! TOOL_SI_HH_
