/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_WINDOW_HH_
# define TOOL_WINDOW_HH_

# include <deque>

# include <trace-tools/tool/iterator.hh>
# include <trace-tools/tool/iterable.hh>

# include "window_fwd.hh"

namespace tool
{

  /// Virtual %types for window. @see types
  template <class I, unsigned Ws, unsigned S, class B>
  struct types< window<I, Ws, S, B> >
  {
    typedef
      std::deque<typename I::value_type>
    value_type;
  };

  /// Virtual %types for window_maker. @see types
  template <class I, unsigned Ws, unsigned S, class B>
  struct types< window_maker<I, Ws, S, B> >
  {
    typedef
      window<I, Ws, S, bottom>
      iterator;
  };

  /**
   * Sliding %window.
   *
   * window provides a sliding %window over a sequence.  It behaves as
   * an iterator which value type is a container for the %window's
   * elements.  Slide the %window using its @link
   * iterator::operator++() operator++() @endlink.  An operator[]() is
   * provided for convenience.
   *
   * @param InputIterator	The sequence's input %iterator type.
   * @param WSize	The window's size (length).
   * @param Step	Number of elements to "eat" when calling increment().
   * @param Bottom	Exact type or bottom (see static_base).
   *
   * @code
   * typedef std::vector<int>          vector_type;
   * typedef vector_type::iterator     iterator_type;
   * typedef window<iterator_type, 2>  window_type;
   *
   * vector_type v;
   * // Fill the vector.
   *
   * window_type w (v, false);
   * window_type end (v, true);
   *
   * while(w != end)
   *   {
   *      std::cout << w[0] << std::endl << w[1] << std::endl << std::endl;
   *      ++w;
   *   }
   * @endcode
   *
   * @see window_maker
   */
  template <class InputIterator,
	    unsigned WSize,
	    unsigned Step = 1,
	    class Bottom = bottom>
  struct window:
    tt_inherit(public iterator, window<InputIterator, WSize, Step>)
  {
    /// This template instantiation's exact type.
    typedef tt_get_exact(Bottom, window<InputIterator, WSize, Step, Bottom>)
								exact_type;

    /// Value type of a %window (@c std::deque<InputIterator::value_type> ).
    typedef tt_type(value_type, exact_type)			value_type;

    /// @link iterable @endlink -compatible constructor.
    template <class Iterable>
    window(Iterable&, bool);

    /// Equality check.  @see iterator.
    bool		equal(const window&) const;

    /// Increment.  @see iterator.
    void		increment();

    /// Indirection (reference).  @see iterator
    const value_type&	get() const;

    /// Indirection (pointer).  @see iterator
    const value_type*	get_ptr() const;

    /// Access to the window's @c i 'th element.
    const typename InputIterator::value_type&
    operator [] (unsigned i) const;

  protected:
    bool		at_end_;///< Is this %window the last one.
    InputIterator	next_;	///< Iterator to the next element to retrieve.
    InputIterator	last_;	///< Iterator past the last element.
    value_type		v_;	///< Current %window value.
  };

  /**
   * Iterable object that builds windows over a sequence.
   *
   * @link window_maker @endlink's main role is to provide a @c
   * begin() and an @c end() method that return windows over
   * a sequence.
   *
   * @see window
   */
  template <class InputIterator,
	    unsigned WSize,
	    unsigned Step = 1,
	    class Bottom = bottom>
  struct window_maker:
    tt_inherit(public iterable,
	       window_maker<InputIterator, WSize, Step, Bottom>)
  {
    /// Construct a window_maker from a sequence.
    window_maker(const InputIterator& first, const InputIterator& last);

  private:
    InputIterator	first_;
    InputIterator	last_;

    friend class window<InputIterator, WSize, Step>;
  };

} // End of namespace tool.

# include "window.hxx"

#endif // ! TOOL_WINDOW_HH_
