/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_ADDR_HXX_
# define WIFI_ADDR_HXX_

# include <cstring>
# include <cassert>
# include <sstream>
# include <iomanip>
# include <stdexcept>
# include <sstream>

# include "addr.hh"

namespace wifi
{

  inline
  addr::addr()
  {
    memset(addr_, 0, sizeof (addr_));
  }

  inline
  addr::addr(const void* a)
  {
    memcpy(addr_, a, sizeof (addr_));
  }

  inline
  addr::addr(const std::string& str)
  {
    std::istringstream	is (str);
    uint8_t		mac[sizeof (addr_)];

    is >> std::hex;
    for (unsigned i = 0; i < sizeof (addr); ++i)
      {
	unsigned	t;

	if (i and is.get() != ':')
	  throw std::invalid_argument ("string is not a MAC address");
	is >> t;
	mac[i] = t;
      }
    memcpy(addr_, mac, sizeof (addr_));
  }

  inline
  bool
  addr::operator < (const addr& rhs) const
  {
    return memcmp(addr_, rhs.addr_, sizeof (addr_)) < 0;
  }

  inline
  bool
  addr::operator == (const addr& rhs) const
  {
    return memcmp(addr_, rhs.addr_, sizeof (addr_)) == 0;
  }

  inline
  bool
  addr::operator != (const addr& rhs) const
  {
    return not operator == (rhs);
  }

  inline
  uint8_t&
  addr::operator [] (unsigned i)
  {
    assert(i < sizeof (addr_));

    return addr_[i];
  }

  inline
  const uint8_t&
  addr::operator [] (unsigned i) const
  {
    assert(i < sizeof (addr_));

    return addr_[i];
  }

  inline
  void
  addr::dump(void *a) const
  {
    memcpy(a, addr_, sizeof (addr_));
  }

  inline
  const addr&
  addr::null()
  {
    static const addr	null;

    return null;
  }

  inline
  const addr&
  addr::broadcast()
  {
    static const uint8_t	raw[6] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
    static const addr		bcast (raw);

    return bcast;
  }

  inline
  std::ostream&
  operator << (std::ostream& os, const addr& a)
  {
    os << std::hex << std::setfill ('0');
    for (unsigned i = 0; i < sizeof (a) - 1; ++i)
      os << std::setw (2) << unsigned (a[i]) << ':';
    return os << std::setw (2) << unsigned (a[sizeof (a) - 1])
	      << std::dec << std::setfill (' ');
  }

  inline
  std::string
  make_string(const addr& a)
  {
    std::ostringstream	os;

    os << a;
    return os.str();
  }

} // End of namespace wifi.

#endif // ! WIFI_ADDR_HXX_
