/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_DISSECTOR_DEFAULT_HOOKS_HH_
# define WIFI_FRAME_DISSECTOR_DEFAULT_HOOKS_HH_

# include <trace-tools/wifi/frame/mgt.hh>
# include <trace-tools/wifi/frame/ctl.hh>
# include <trace-tools/wifi/frame/data.hh>
# include <trace-tools/wifi/frame/dissector/status.hh>

namespace wifi
{

  namespace frame
  {
    /**
     * Default (empty) hooks to be used with dissector.
     *
     * The dissector class template parses IEEE 802.11 %frames and
     * calls hooks at different stages of this parsing.  This class
     * provides all the hook functions a dissector may need.  These
     * implementations are empty, @e i.e. they do nothing when called.
     *
     * Implementing hook functions is a tedious work, and it is
     * unlikely you need all the hooks that dissector requires.  This
     * class therefore aims at making the task easier.  Just subclass
     * it so you will not have to provide implementations of hooks you
     * do not need.  You may then just override the hooks you need.
     *
     * Hooks have consistent prototypes.  Their first two arguments
     * are always a pointer to the %frame, and the number of available
     * bytes in the %frame.  When relevant (that is, when the parser
     * state may vary through different calls to this hook) a third
     * argument provides the parser state.  Sometimes there is also
     * extra arguments such as the address that has just been parsed,
     * the %frame sequence number, etc.
     */
    struct dissector_default_hooks
    {
      /// @link dissector Dissector @endlink status.
      typedef dissector_status::status	status;

    protected:

      //        ******************* //
      /// @name Invalid frame hooks
      //        ******************* //
      //@{

      void	invalid_type_or_subtype_hook(const void*, size_t, status);
      void	invalid_type_hook(const void*, size_t);
      void	invalid_subtype_hook(const void*, size_t, status);
      void	invalid_management_subtype_hook(const mgt::header*, size_t);
      void	invalid_control_subtype_hook(const void*, size_t);
      void	invalid_data_subtype_hook(const void*, size_t, status);

      //@}


      //        *********** //
      /// @name Basic hooks
      //        *********** //
      //@{


      /// Called before starting the parsing.
      void	frame_hook(const void*, size_t);

      /**
       * An address field has just been parsed.
       *
       * The third argument is the address' position (@e i.e. 1, 2, 3,
       * or 4) and the last argument is the address itself.
       */
      void	addr_hook(const void*, size_t, unsigned, const addr&);

      /**
       * The sequence control field has just been parsed.
       *
       * The third argument is the fragment number.  The fourth one is
       * the sequence number.
       */
      void	seq_ctl_hook(const void*, size_t, unsigned, unsigned);

      /// End of %frame reached.
      void	end_of_frame_hook(const void*, size_t, status);

      /// Called before starting to parse a management %frame.
      void	management_hook(const mgt::header*, size_t);

      /// Called before starting to parse a control %frame.
      void	control_hook(const void*, size_t);

      /// Called before starting to parse a %data %frame.
      void	data_hook(const void*, size_t);

      //@}


      //        *************** //
      /// @name Truncation hook
      //        *************** //
      //@{

      void	truncated_frame_hook(const void*, size_t, status);
      void	truncated_unknown_frame_hook(const void*, size_t);
      void	truncated_management_frame_hook(const mgt::header*, size_t,
						status);
      void	truncated_ps_poll_frame_hook(const ctl::ps_poll::header*,
					     size_t, status);
      void	truncated_rts_frame_hook(const ctl::rts::header*, size_t,
					 status);
      void	truncated_cf_end_frame_hook(const ctl::cf_end::header*, size_t,
					    status);
      void	truncated_cf_end_cf_ack_frame_hook(const ctl::cf_end_cf_ack::
						   header*, size_t, status);
      void	truncated_cts_frame_hook(const ctl::cts::header*, size_t);
      void	truncated_ack_frame_hook(const ctl::ack::header*, size_t);
      void	truncated_data_frame_hook(const void*, size_t, status);

      // @}

      //        ********************** //
      /// @name Management frame hooks
      //        ********************** //
      //@{

      // Field hooks.
      void	management_addr1_hook(const mgt::header*, size_t);
      void	management_addr2_hook(const mgt::header*, size_t);
      void	management_addr3_hook(const mgt::header*, size_t);
      void	management_seq_ctl_hook(const mgt::header*, size_t,
					unsigned, unsigned);

      // Inidividual frame subtypes.
      void	beacon_hook(const mgt::header*, size_t);
      void	assoc_req_hook(const mgt::header*, size_t);
      void	assoc_resp_hook(const mgt::header*, size_t);
      void	reassoc_req_hook(const mgt::header*, size_t);
      void	reassoc_resp_hook(const mgt::header*, size_t);
      void	probe_req_hook(const mgt::header*, size_t);
      void	probe_resp_hook(const mgt::header*, size_t);
      void	atim_hook(const mgt::header*, size_t);
      void	disassoc_hook(const mgt::header*, size_t);
      void	auth_hook(const mgt::header*, size_t);
      void	deauth_hook(const mgt::header*, size_t);

      //@}


      //        ******************* //
      /// @name Control frame hooks
      //        ******************* //
      //@{

      // Individual frame subtypes.
      void	ps_poll_hook(const ctl::ps_poll::header*, size_t);
      void	rts_hook(const ctl::rts::header*, size_t);
      void	cf_end_hook(const ctl::cf_end::header*, size_t);
      void	cf_end_cf_ack_hook(const ctl::cf_end_cf_ack::header*, size_t);
      void	cts_hook(const ctl::cts::header*, size_t);
      void	ack_hook(const ctl::ack::header*, size_t);

      // Field hooks //

      // PS Poll.
      void	ps_poll_addr1_hook(const ctl::ps_poll::header*, size_t);
      void	ps_poll_addr2_hook(const ctl::ps_poll::header*, size_t);

      // RTS.
      void	rts_addr1_hook(const ctl::rts::header*, size_t);
      void	rts_addr2_hook(const ctl::rts::header*, size_t);

      // CF-End.
      void	cf_end_addr1_hook(const ctl::cf_end::header*, size_t);
      void	cf_end_addr2_hook(const ctl::cf_end::header*, size_t);

      // CF-End + CF-Ack.
      void cf_end_cf_ack_addr1_hook(const ctl::cf_end_cf_ack::header*, size_t);
      void cf_end_cf_ack_addr2_hook(const ctl::cf_end_cf_ack::header*, size_t);

      // CTS.
      void	cts_addr1_hook(const ctl::cts::header*, size_t);

      // Ack.
      void	ack_addr1_hook(const ctl::ack::header*, size_t);

      //@}

      //        **************** //
      /// @name Data frame hooks
      //        **************** //
      //@{

      // Field hooks.
      void	data_addr1_hook(const void*, size_t);
      void	data_addr2_hook(const void*, size_t);
      void	data_addr3_hook(const void*, size_t);
      void	data_addr4_hook(const void*, size_t);
      void	data_seq_ctl_hook(const void*, size_t, unsigned, unsigned);

      // Individual frame subtypes.
      void	data_only_hook(const void*, size_t, status);
      void	data_cf_ack_hook(const void*, size_t, status);
      void	data_cf_poll_hook(const void*, size_t, status);
      void	data_cf_ack_cf_poll_hook(const void*, size_t, status);
      void	null_hook(const void*, size_t, status);
      void	cf_ack_hook(const void*, size_t, status);
      void	cf_poll_hook(const void*, size_t, status);
      void	cf_ack_cf_poll_hook(const void*, size_t, status);

      void	within_ibss_hook(const data::within_ibss::header*, size_t);
      void	from_ds_hook(const data::from_ds::header*, size_t);
      void	to_ds_hook(const data::to_ds::header*, size_t);
      void	ap_to_ap_hook(const data::ap_to_ap::header*, size_t);

      //@}

    };

  } // End of namespace wifi::frame.

} // End of namespace wifi.

# include "default_hooks.hxx"

#endif // ! WIFI_FRAME_DISSECTOR_DEFAULT_HOOKS_HH_
