/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_DISSECTOR_DISSECTOR_HH_
# define WIFI_FRAME_DISSECTOR_DISSECTOR_HH_

# include <cstdlib>

# include <trace-tools/wifi/frame/mgt.hh>
# include <trace-tools/wifi/frame/dissector/status.hh>

namespace wifi
{

  namespace frame
  {

    struct dissector_default_hooks;

    /**
     * IEEE 802.11 %frame parser.
     *
     * This class template provides mechanisms to parse IEEE 802.11
     * frames.  Its template parameter should be a class that defines
     * several hooks to be called at different parsing stages.
     *
     * There is basically two modes of operation : triggering the
     * parsing implicitly at object creation, or explicitly by
     * calling operator()().
     *
     * @code
     * const void*     frame = // ...
     * const unsigned  caplen = // ...
     *
     * my_hooks  hooks1 = // ...
     * my_hooks  hooks2 = // ...
     *
     * dissector<my_hooks>  d1 (hooks1);                // No parsing done yet.
     * dissector<my_hooks>  d2 (frame, caplen, hooks2); // Implicitely parse
     *                                                  // frame.
     *
     * d1(frame, caplen); // Do parse frame explicitely.
     * @endcode
     *
     * There is many different hooks.  Look @link
     * dissector_default_hooks @endlink's documentation for a list.
     * Although subclassing dissector_default_hooks is a good idea in
     * order to implement your set of hooks, this is not mandatory.
     *
     * The status() method allows you to get the parser's status after
     * the parsing.  When this is relevant, this state is provided to
     * hook functions as argument.
     *
     * @see dissector_default_hooks, dissector_status::status
     */
    template <class Hooks = dissector_default_hooks>
    struct dissector: public Hooks
    {
      /// Construct a %dissector but do not trigger parsing yet.
      dissector(const Hooks& = Hooks ());

      /**
       * Construct a %dissector and parse a %frame.
       *
       * @param[in]  frm	Pointer to the %frame.
       * @param[in]  caplen	Available bytes in the %frame.  The %frame may
       *			be truncated.  In such case, the parsing ends
       *			with the parsing in an intermediary state.
       * @param[in] hooks	The hooks to use.
       *
       * @see status()
       */
      dissector(const void* frm, size_t caplen, const Hooks& hooks = Hooks ());

      /// Get the current parser state.  @see dissector_status::status
      dissector_status::status	status() const;

      /**
       * Do parse a %frame.
       *
       * @param[in]  frm	Pointer to the %frame.
       * @param[in]  caplen	Available bytes in the %frame.  The %frame may
       *			be truncated.  In such case, the parsing ends
       *			with the parsing in an intermediary state.
       *
       * @see status()
       */
      void			operator () (const void* frm, size_t caplen);

      /// @name Hooks accessors
      //@{
      Hooks&			hooks();
      const Hooks&		hooks() const;
      //@}

    private:

      /// Needed number of captured bytes to analyze various fields.
      enum length
      {
	control		= 2,
	duration	= 4,
	addr1		= 10,
	addr2		= 16,
	addr3		= 22,
	seqctl		= 24,
	addr4		= 30
      };

      void			management_frame(const void*, size_t);
      void			management_frame_dispatch(const mgt::header*,
							  size_t);

      void			control_frame(const void*, size_t);

      void			data_frame(const void*, size_t);
      void			data_frame_dispatch(const void*, size_t);

      dissector_status::status	status_;
    };

  } // End of namespace frame.

} // End of namespace wifi.

# include "dissector.hxx"

#endif // ! WIFI_FRAME_DISSECTOR_DISSECTOR_HH_
