/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#include <iostream>
#include <stdexcept>

#include <boost/date_time/posix_time/posix_time.hpp>

#include <trace-tools/pcap/descriptor.hh>

#include "config.hh"

namespace
{

  struct context
  {
    typedef boost::posix_time::time_period	time_period;

    context(pcapxx::dumper& dump, const time_period& period): dump_ (dump),
							      period_ (period)
    {
    }

    void
    got_frame(const pcapxx::pkthdr*	h,
	      const void*		p)
    {
      using namespace boost::gregorian;
      using namespace boost::posix_time;

      const ptime toa (date(1970, Jan, 01),
		       seconds (h->ts.tv_sec) + microseconds (h->ts.tv_usec));

      if (period_.contains(toa))
	dump_(h, p);
    }

  private:
    pcapxx::dumper&	dump_;
    const time_period&	period_;
  };

  void
  got_frame(void*			c,
	    const pcapxx::pkthdr*	h,
	    const void*			p)
  {
    static_cast<context*> (c)->got_frame(h, p);
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  if (argc != 3)
    {
      std::cerr << "Usage:\t"  << argv[0] << " <start-date> <stop-date>"
		<< std::endl;
      return EXIT_FAILURE;
    }

  try
    {
      using boost::posix_time::ptime;
      using boost::posix_time::time_period;

      const ptime	start = boost::posix_time::time_from_string (argv[1]);
      const ptime	stop = boost::posix_time::time_from_string(argv[2]);
      const time_period		period (start, stop);
      pcapxx::descriptor<>	desc ("-");
      pcapxx::dumper		dump = desc.dumper("-");
      context			ctx (dump, period);

      desc.loop<got_frame>(static_cast<void*> (&ctx), -1);
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
